/** @file sysc_synchronization.c

Copyright (c) 2021 Microchip Technology Inc. All Rights Reserved.

Description: Template sysCore synchronization source file

*/

#ifndef SC_DISABLE_SYNC_API

#include "psysc_platform.h"
#include "psysc_types.h"

/** Define the following specific to the platform:
 * - COND_VARIABLE and PCOND_VARIABLE
 * - MUTEX_LOCK and PMUTEX_LOCK
 */

/** An example of COND_VARIABLE and PCOND_VARIABLE definitions is shown below. */
struct _COND_VARIABLE
{
   /** Define other members specific to the platform. */
   void *variable;
   SA_BOOL initialized;
};
typedef struct _COND_VARIABLE COND_VARIABLE;
typedef struct _COND_VARIABLE *PCOND_VARIABLE;

/** An example of MUTEX_LOCK and PMUTEX_LOCK definitions is shown below. */
struct _MUTEX_LOCK
{
   /** Define other members specific to the platform. */
   void *lock;
   SA_BOOL initialized;
   COND_VARIABLE cond;
   SA_DWORD tid;
   long int lock_depth;
};
typedef struct _MUTEX_LOCK MUTEX_LOCK;
typedef struct _MUTEX_LOCK *PMUTEX_LOCK;

SA_DWORD SC_GetCurrentThreadID();

/* *******************
 * Condition Variables
 */

static SA_BOOL _SC_CondVariableInitialize(PCOND_VARIABLE cond)
{
   SA_BOOL success = kFalse;
   if (cond)
   {
      /*
      success = (pthread_cond_init(&cond->variable, 0) == 0 ? kTrue : kFalse);
      */
      cond->initialized = kTrue;
   }
   return success;
}

static SA_BOOL _SC_CondVariableDestroy(PCOND_VARIABLE cond)
{
   SA_BOOL success = kFalse;
   if (cond)
   {
      /*
      success = (pthread_cond_destroy(&cond->variable) == 0 ? kTrue : kFalse);
      */
      cond->initialized = kFalse;
   }
   return success;
}

SA_BOOL _SC_CondVariableCreate(PCOND_VARIABLE *cond)
{
   SA_BOOL success = kFalse;
   if (cond)
   {
      if ((*cond = (PCOND_VARIABLE)SA_malloc(sizeof(COND_VARIABLE))) != 0)
      {
         success = _SC_CondVariableInitialize(*cond);

         if (!success)
         {
            SA_free(*cond);
            *cond = 0;
         }

      }
   }
   return success;
}

void _SC_CondVariableDelete(PCOND_VARIABLE cond)
{
   if (cond)
   {
      (void)_SC_CondVariableDestroy(cond);
      SA_free(cond);
   }
}

SA_BOOL _SC_CondVariableWait(PCOND_VARIABLE cond, PMUTEX_LOCK mutex)
{
   SA_BOOL success = kFalse;

   if (cond && mutex && cond->initialized && mutex->initialized)
   {
      /*
      success = (pthread_cond_wait(&cond->variable, &mutex->lock) == 0 ? kTrue : kFalse);
      */
   }

   return success;
}

SA_BOOL _SC_CondVariableSignal(PCOND_VARIABLE cond)
{
   SA_BOOL success = kFalse;
   if (cond && cond->initialized)
   {
      /*
      success = (pthread_cond_signal(&cond->variable) == 0 ? kTrue : kFalse);
      */
   }
   return success;
}

SA_BOOL _SC_CondVariableBroadcast(PCOND_VARIABLE cond)
{
   SA_BOOL success = kFalse;
   if (cond && cond->initialized)
   {
      /*
      success = (pthread_cond_broadcast(&cond->variable) == 0 ? kTrue : kFalse);
      */
   }
   return success;
}

/* *****
 * Mutex
 */

static SA_BOOL _SC_MutexInitialize(PMUTEX_LOCK mutex)
{
   SA_BOOL success = kFalse;
   if (mutex)
   {
      mutex->tid = 0;
      mutex->lock_depth = 0;

      /*
      success = (pthread_mutex_init(&mutex->lock, 0) == 0 ? kTrue : kFalse);
      */

      if (success)
      {
         success = _SC_CondVariableInitialize(&mutex->cond);
      }

      mutex->initialized = kTrue;
   }
   return success;
}

SA_BOOL _SC_MutexCreate(PMUTEX_LOCK *mutex)
{
   SA_BOOL success = kFalse;
   if ((*mutex = (PMUTEX_LOCK)SA_malloc(sizeof(MUTEX_LOCK))) != 0)
   {
      success = _SC_MutexInitialize(*mutex);

      if (!success)
      {
         SA_free(*mutex);
         *mutex = 0;
      }

   }
   return success;
}

static SA_BOOL _SC_MutexDestroy(PMUTEX_LOCK mutex)
{
   SA_BOOL success = kFalse;
   if (mutex)
   {
      /*
      success = (pthread_mutex_destroy(&mutex->lock) == 0 ? kTrue : kFalse);
      */

      _SC_CondVariableDestroy(&mutex->cond);
      mutex->tid = 0;
      mutex->lock_depth = 0;
      mutex->initialized = kFalse;
   }
   return success;
}

void _SC_MutexDelete(PMUTEX_LOCK mutex)
{
   if (mutex)
   {
      (void)_SC_MutexDestroy(mutex);
      SA_free(mutex);
   }
}

SA_BOOL _SC_MutexLock(PMUTEX_LOCK mutex)
{
   int err = 1;
   if (mutex && mutex->initialized)
   {
      /*
      err = pthread_mutex_lock(&mutex->lock);
      */
   }
   return (err == 0 ? kTrue : kFalse);
}

SA_BOOL _SC_MutexUnlock(PMUTEX_LOCK mutex)
{
   int err = 1;
   if (mutex && mutex->initialized)
   {
      /*
      err = pthread_mutex_unlock(&mutex->lock);
      */
   }
   return (err == 0 ? kTrue : kFalse);
}

/* ***************
 * Recursive Mutex
 */

SA_BOOL _SC_RecursiveMutexLock(PMUTEX_LOCK mutex)
{
   SA_BOOL success = kFalse;
   if (mutex && mutex->initialized)
   {
      SA_DWORD tid = SC_GetCurrentThreadID();
      success = _SC_MutexLock(mutex);

      if (mutex->lock_depth > 0 && mutex->tid == tid)
      {
         mutex->lock_depth++;
      }
      else
      {
         while (mutex->lock_depth > 0)
         {
            success = ((_SC_CondVariableWait(&mutex->cond, mutex) && success) ? kTrue : kFalse);
         }
         mutex->tid = tid;
         mutex->lock_depth = 1;
      }
      success = ((_SC_MutexUnlock(mutex) && success) ? kTrue : kFalse);

   }
   return success;
}

SA_BOOL _SC_RecursiveMutexUnlock(PMUTEX_LOCK mutex)
{
   SA_BOOL success = kFalse;
   if (mutex && mutex->initialized)
   {
      SA_DWORD tid = SC_GetCurrentThreadID();
      success = _SC_MutexLock(mutex);

      if (mutex->lock_depth > 0 && mutex->tid == tid)
      {
         mutex->lock_depth--;
         if (mutex->lock_depth <= 0)
         {
            mutex->lock_depth = 0;
            success = ((_SC_CondVariableSignal(&mutex->cond) && success) ? kTrue : kFalse);
         }
      }
      success = ((_SC_MutexUnlock(mutex) && success) ? kTrue : kFalse);

   }
   return success;
}

#endif /* SC_DISABLE_SYNC_API */

/* ***************
 * Event Flags
 */

/* sysCore clients need to define the following specific to the platform:
 * - EVENT_FLAGS and PEVENT_FLAGS
 */

/** An example EVENT_FLAGS and PEVENT_FLAGS definitions is shown below. */
struct _EVENT_FLAGS
{
   /** Define other members specific to the platform. */
   SA_DWORD flags;
   MUTEX_LOCK lock;
   SA_BOOL initialized;
};
typedef struct _EVENT_FLAGS EVENT_FLAGS;
typedef struct _EVENT_FLAGS *PEVENT_FLAGS;

#ifndef SC_DISABLE_EVENT_FLAGS

SA_BOOL _SC_EventFlagsInitialize(PEVENT_FLAGS event_flags)
{
   SA_BOOL success = kFalse;
   success = _SC_MutexInitialize(&event_flags->lock);
#if _BullseyeCoverage
#pragma BullseyeCoverage off
#endif
   if (success)
   {
      event_flags->initialized = kTrue;
   }
#if _BullseyeCoverage
#pragma BullseyeCoverage on
#endif
   return success;
}

SA_BOOL _SC_EventFlagsCreate(PEVENT_FLAGS *event_flags)
{
   SA_BOOL success = kFalse;
   if (event_flags &&
      (*event_flags = (PEVENT_FLAGS)SA_malloc(sizeof(EVENT_FLAGS))) != 0)
   {
      success = _SC_EventFlagsInitialize(*event_flags);
#if _BullseyeCoverage
#pragma BullseyeCoverage off
#endif
      if (!success)
      {
         SA_free(*event_flags);
         *event_flags = 0;
      }
#if _BullseyeCoverage
#pragma BullseyeCoverage on
#endif
   }
   return success;
}

static SA_BOOL _SC_EventFlagsDestroy(PEVENT_FLAGS event_flags)
{
   return _SC_MutexDestroy(&event_flags->lock);
}

SA_BOOL _SC_EventFlagsDelete(PEVENT_FLAGS event_flags)
{
   SA_BOOL success = kFalse;
   if (event_flags)
   {
      success = _SC_EventFlagsDestroy(event_flags);
      SA_free(event_flags);
   }
   return success;
}

#define kEventFlagsSetOR        0x00000000
#define kEventFlagsSetAND       0x00000002
SA_BOOL _SC_EventFlagsSet(PEVENT_FLAGS event_flags, SA_DWORD flags_to_set, SA_DWORD set_options)
{
   SA_BOOL success = kFalse;
   if (event_flags && event_flags->initialized &&
      (set_options == kEventFlagsSetOR
       || set_options == kEventFlagsSetAND
       ))
   {
      success = _SC_MutexLock(&event_flags->lock);

      if (set_options & kEventFlagsSetAND)
         event_flags->flags &= flags_to_set;
      else /* if (set_options == kEventFlagsSetOR) */
         event_flags->flags |= flags_to_set;

      success = ((_SC_CondVariableBroadcast(&event_flags->lock.cond) && success) ? kTrue : kFalse);
      success = ((_SC_MutexUnlock(&event_flags->lock) && success) ? kTrue : kFalse);
   }
   return success;
}

#define kEventFlagsGetOR         0x00000001
#define kEventFlagsGetORClear    0x00000002
#define kEventFlagsGetAND        0x00000004
#define kEventFlagsGetANDClear   0x00000008
#define kEventFlagsMaskCtrlFlags 0x0000000F

#define kEventFlagsGetWait       0x80000000
#define kEventFlagsGetNoWait     0x40000000
#define kEventFlagsGetBlockFlags 0xC0000000
SA_BOOL _SC_EventFlagsGet(PEVENT_FLAGS event_flags, SA_DWORD requested_flags, SA_DWORD get_options, SA_DWORD *actual_flags)
{
   SA_BOOL success = kFalse;
   SA_DWORD mask_ctrl_flags = (get_options & kEventFlagsMaskCtrlFlags);
   SA_DWORD block_flags = (get_options & kEventFlagsGetBlockFlags);
   if (event_flags &&
         event_flags->initialized &&
         actual_flags &&
         (
          mask_ctrl_flags == kEventFlagsGetOR
          || mask_ctrl_flags == kEventFlagsGetORClear
          || mask_ctrl_flags == kEventFlagsGetAND
          || mask_ctrl_flags == kEventFlagsGetANDClear
         ) &&
         (
          block_flags == kEventFlagsGetWait
          || block_flags == kEventFlagsGetNoWait
         ))
   {
      success = _SC_MutexLock(&event_flags->lock);
      while (!(
               /* AND/Clear and all requested flags set */
               ((mask_ctrl_flags == kEventFlagsGetAND || mask_ctrl_flags == kEventFlagsGetANDClear) &&
                ((event_flags->flags & requested_flags) == requested_flags))
               /* OR/Clear and any requested flags set */
               ||
               ((mask_ctrl_flags == kEventFlagsGetOR || mask_ctrl_flags == kEventFlagsGetORClear) &&
                (event_flags->flags & requested_flags))
              )
             && block_flags == kEventFlagsGetWait)
      {
         success = ((_SC_CondVariableWait(&event_flags->lock.cond, &event_flags->lock) && success) ? kTrue : kFalse);
      }

      *actual_flags = (event_flags->flags & requested_flags);

      if (mask_ctrl_flags == kEventFlagsGetORClear ||
         mask_ctrl_flags == kEventFlagsGetANDClear)
      {
         event_flags->flags &= ~requested_flags;
      }

      success = ((_SC_MutexUnlock(&event_flags->lock) && success) ? kTrue : kFalse);
   }
   return success;
}

#endif /* SC_DISABLE_EVENT_FLAGS */

